"""
OdinWeb.Django API
~~~~~~~~~~~~~~~~~~

Django implementation of the OdinWeb API interface.

"""
from __future__ import absolute_import

import collections
from django.conf import settings
from django.conf.urls import url
from django.http import HttpRequest, HttpResponse, HttpResponseNotAllowed
from django.views.decorators.csrf import csrf_exempt

from odinweb.containers import ApiInterfaceBase
from odinweb.constants import Type, Method, PATH_STRING_RE
from odinweb.data_structures import PathParam


TYPE_MAP = {
    Type.Integer: r'\d+',
    Type.Long: r'\d+',
    Type.Float:  r'[.\d]+',
    Type.Double:  r'[.\d]+',
    Type.String: PATH_STRING_RE,
    Type.Byte: '',
    Type.Binary: '',
    Type.Boolean: r'0|1|true|false|yes|no|on|off',
    Type.Date: r'\d{4}-\d{2}-\d{2}',
    Type.Time: r'\d{2}:\d{2}',
    Type.DateTime: r'[-:\d]+',
    Type.Password: PATH_STRING_RE,
}


class RequestProxy(object):
    def __init__(self, r):
        # type: (HttpRequest) -> None
        self.GET = r.GET
        self.POST = r.POST
        self.headers = r.META
        # self.session = r.
        self.request = r

        try:
            method = Method[r.method]
        except KeyError:
            method = None
        self.method = method

    @property
    def body(self):
        return self.request.body

    @property
    def host(self):
        return self.request.get_host()


class Api(ApiInterfaceBase):
    """
    An API base for Django::

        from odinweb.django import Api

        url_patterns = Api(
            ApiVersion(
                UserApi(),
                version='v1
            )
        ).urls()

    """
    def __init__(self, *containers, **options):
        options.setdefault('debug_enabled', settings.DEBUG)
        super(Api, self).__init__(*containers, **options)

    @staticmethod
    def node_formatter(path_node):
        # type: (PathParam) -> str
        """
        Format a node to be consumable by the `UrlPath.parse`.
        """
        node_type = TYPE_MAP.get(path_node.type, '\w+')  # Generic string default
        return r"(?P<{}>{})".format(path_node.name, node_type)

    def _bound_callback(self, methods):
        @csrf_exempt
        def callback(request, **path_args):
            request = RequestProxy(request)

            operation = methods.get(request.method)
            if operation:
                # Dispatch request to framework
                response = self.dispatch(operation, request, **path_args)

                # Build Django response
                django_response = HttpResponse(response.body, status=response.status)
                for key, value in response.headers.items():
                    django_response[key] = value
                return django_response

            else:
                return HttpResponseNotAllowed(m.value for m in methods)
        return callback

    def urls(self):
        """
        URLs to be integrated into Django url_patterns lists.
        """
        # Build URLs
        results = []
        for url_path, methods in self.op_paths(collate_methods=True).items():
            # Generate path and apply regex wrapping.
            path = r'^{}$'.format(
                url_path.format(self.node_formatter)[1:]
            )

            # Let framework handle unknown method.
            # for method in tuple(m.value for m in operation.methods):
            results.append(url(path, self._bound_callback(methods)))
        return results



import odin

from odinweb import api, doc
from odinweb.django.api.models import ModelResourceApi, Listing, Create, Detail, Update, Delete
from odinweb.swagger import SwaggerSpec

from my_app.models import Group


class User(odin.Resource):
    """
    User resource
    """
    id = odin.IntegerField()
    username = odin.StringField()
    name = odin.StringField()
    email = odin.EmailField()
    role = odin.StringField(choices=('a', 'b', 'c'))

USERS = [
    User(1, 'pimpstar24', 'Bender', 'Rodreges', 'bender@ilovebender.com'),
    User(2, 'zoidberg', 'Zoidberg', '', 'zoidberg@freemail.web'),
    User(3, 'amylove79', 'Amy', 'Wong', 'awong79@marslink.web'),
]
USER_ID = len(USERS)


class UserApi(api.ResourceApi):
    resource = User
    tags = ['user']

    @doc.deprecated
    @api.collection(path='find', methods=api.Method.POST)
    def operation_test(self, request):
        pass

    @api.listing
    def get_user_list(self, request, offset, limit):
        return USERS[offset:offset+limit], len(USERS)

    @api.create
    def create_user(self, request, user):
        global USER_ID

        # Add user to list
        USER_ID += 1
        user.id = USER_ID
        USERS.append(user)

        return user

    @api.detail
    def get_user(self, request, resource_id):
        """
        Get a user object
        """
        for user in USERS:
            if user.id == resource_id:
                return user

        raise api.HttpError(api.HTTPStatus.NOT_FOUND)

    @api.update
    def update_user(self, request, user, resource_id):
        return user

    @api.patch
    def patch_user(self, request, user, resource_id):
        return user

    @api.delete
    def delete_user(self, request, resource_id):
        for idx, user in enumerate(USERS):
            if user.id == resource_id:
                USERS.remove(user)
                return

        raise api.HttpError(api.HTTPStatus.NOT_FOUND)


class GroupApi(Listing, Create, Detail, Update, Delete, ModelResourceApi):
    """
    Endpoint to managing groups.
    """
    model = Group
    tags = ['group']


sample_api = api.ApiCollection(name='sample')


@sample_api.operation('foo/bar')
def sample(request):
    return {}


api_v1 = api.ApiVersion(
    SwaggerSpec("Flask Example Swaggerspec", enable_ui=True),
    sample_api,
    UserApi(),
    GroupApi(),
)



from odinweb import _compat
from odinweb.exceptions import PermissionDenied, AccessDenied


class LoginRequired(object):
    """
    A login is required for this API.
    """
    def pre_dispatch(self, request, _):
        if not request.request.user.is_authenticated:
            raise PermissionDenied(message="Login is required")


class PermissionRequired(object):
    """
    Certain permission is required for this API.
    """
    def __init__(self, perm):
        if isinstance(perm, _compat.string_types):
            self.perms = (perm, )
        else:
            self.perms = perm

    def pre_dispatch(self, request, _):
        if not request.request.user.has_perms(self.perms):
            raise AccessDenied(message="User doesn't have required permissions.")



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django.shortcuts import get_object_or_404
from odinweb.helpers import create_response

from odin import registration
from odin.exceptions import CodecDecodeError

from odinweb import api
from odinweb.django.models import model_resource_factory
from odinweb.constants import HTTPStatus
from odinweb.exceptions import HttpError


class ModelResourceApi(api.ResourceApi):
    """
    Provides an API for working with Django models.

    This API assumes that mappings have been defined between between the model and a suitable resource.

    """
    # Model this API deals with
    model = None
    # Model field to use for single model queries.
    model_id_field = 'pk'
    # Mapping to use for mapping to model
    to_model_mapping = None
    # Mapping to use for mapping to resource
    to_resource_mapping = None

    def __init__(self, *args, **kwargs):
        assert self.model, "A model has not been provided."

        if self.resource is None:
            self.resource = model_resource_factory(self.model)

        super(ModelResourceApi, self).__init__(*args, **kwargs)

        # Attempt to resolve mappings
        if self.to_model_mapping is None:
            self.to_model_mapping = registration.get_mapping(self.resource, self.model)
        if self.to_resource_mapping is None:
            self.to_resource_mapping = registration.get_mapping(self.model, self.resource)

    def get_queryset(self, request):
        return self.model.objects.all()

    def get_instance(self, request, resource_id):
        return get_object_or_404(self.get_queryset(request), **{
            self.model_id_field: resource_id
        })

    def update_instance_from_body(self, request, instance, resource=None, ignore_fields=('id', 'pk')):
        """
        Get a resource that merges an instance and the request body.
        :param request:
        :param instance:
        :param resource:
        :param ignore_fields:
        :return:

        """
        resource = resource or self.resource

        try:
            body = self.decode_body(request)
        except UnicodeDecodeError as ude:
            raise HttpError(HTTPStatus.BAD_REQUEST, 100, "Unable to decode request body.", str(ude))

        try:
            resource = request.request_codec.loads(body, resource=resource, full_clean=False,
                                                   default_to_not_supplied=True)
        except ValueError as ve:
            raise HttpError(HTTPStatus.BAD_REQUEST, 98, "Unable to load resource.", str(ve))
        except CodecDecodeError as cde:
            raise HttpError(HTTPStatus.BAD_REQUEST, 96, "Unable to decode body.", str(cde))

        # Update only the supplied fields
        self.to_model_mapping(resource).update(instance, ignore_fields=ignore_fields)

        return resource

    def save_model(self, request, instance, is_new=False):
        instance.save()
        return instance


class Listing(ModelResourceApi):
    """
    Mixin that provides a basic paged listing response.
    """
    @api.listing
    def object_list(self, request, limit, offset):
        queryset = self.get_queryset(request)
        page = queryset[offset:offset+limit]
        return self.to_resource_mapping.apply(page), len(queryset)


class Create(ModelResourceApi):
    """
    Mixin that provides a basic creation method.
    """
    @api.create
    def object_create(self, request, resource):
        instance = self.to_model_mapping.apply(resource)
        instance.id = None
        self.save_model(request, instance, True)
        return create_response(request, self.to_resource_mapping.apply(instance), HTTPStatus.CREATED)


class Detail(ModelResourceApi):
    """
    Mixin that provides a basic full detail method.
    """
    @api.detail
    def object_detail(self, request, resource_id):
        instance = self.get_instance(request, resource_id)
        return self.to_resource_mapping.apply(instance)


class Update(ModelResourceApi):
    """
    Mixin that provides a basic model update method.
    """
    @api.update
    def object_update(self, request, resource, resource_id):
        instance = self.get_instance(request, resource_id)
        self.to_model_mapping(resource).update(instance, ignore_fields=('id', 'pk'))
        self.save_model(request, instance, False)
        return self.to_resource_mapping.apply(instance)

#
# class PatchMixin(ModelResourceApi):
#     """
#     Mixin that provides a basic model update method.
#     """
#     @api.patch
#     def object_update(self, request, resource, resource_id):
#         instance = self.get_instance(request, resource_id)
#         self.update_instance_from_body(request, instance)
#         self.save_model(request, instance, False)
#         return self.to_resource_mapping.apply(instance)


class Delete(ModelResourceApi):
    """
    Mixin that provides a basic delete method.
    """
    @api.delete
    def object_delete(self, request, resource_id):
        self.get_instance(request, resource_id).delete()
